# Starting code for Assignment 5

#-------------------------------------------------------------------
# 3-layer neural network with 2 inputs, 2 hidden units, and 1 output
#
#                           output
#                             / \
#              output_w1 --->/   \<--- output_w2
#                           /     \
#                       hidden1 hidden2
#                          |\     /|
#           hidden1_w1 --->| \   / |<--- hidden2_w2
#           hidden1_w2 ----|->\ /<-|---- hidden2_w1
#                          |   X   |
#                          |  / \  |
#                        input1 input2
#
#-------------------------------------------------------------------

import random, math
import NNinspector  # automated tester program

def sigmoid(z):
    return 1 / (1 + math.exp(-z))

class NeuralNetwork:

    def __init__(self):
        # unit activations
        self.output = 0
        self.hidden1 = 0
        self.hidden2 = 0
        # output unit weights
        self.output_bias = random.uniform(-0.1, +0.1)
        self.output_w1 = random.uniform(-0.1, +0.1)
        self.output_w2 = random.uniform(-0.1, +0.1)
        # hidden1 unit weights
        self.hidden1_bias = random.uniform(-0.1, +0.1)
        self.hidden1_w1 = random.uniform(-0.1, +0.1)
        self.hidden1_w2 = random.uniform(-0.1, +0.1)
        # hidden2 unit weights
        self.hidden2_bias = random.uniform(-0.1, +0.1)
        self.hidden2_w1 = random.uniform(-0.1, +0.1)
        self.hidden2_w2 = random.uniform(-0.1, +0.1)
        # learning parameters
        self.tolerance = 0.1
        self.epoch_limit = 50000
        self.learning_rate = 0.1
        # for use in Part 2
        self.momentum = 0

    def __str__(self):
        s = "           bias     w1      w2\n"
        s += "output   {:+.4f} {:+.4f} {:+.4f}\n".format(
            self.output_bias, self.output_w1, self.output_w2)
        s += "hidden1  {:+.4f} {:+.4f} {:+.4f}\n".format(
            self.hidden1_bias, self.hidden1_w1, self.hidden1_w2)
        s += "hidden2  {:+.4f} {:+.4f} {:+.4f}".format(
            self.hidden2_bias, self.hidden2_w1, self.hidden2_w2)
        return s
        
    def set_weights(self, weight_list):
        assert type(weight_list) is list and len(weight_list) == 9, \
            "set_weights requires a list of 9 weight/bias values"
        # weights in weight_list must be given in the order below:
        self.output_bias  = weight_list[0]
        self.output_w1    = weight_list[1]
        self.output_w2    = weight_list[2]
        self.hidden1_bias = weight_list[3]
        self.hidden1_w1   = weight_list[4]
        self.hidden1_w2   = weight_list[5]
        self.hidden2_bias = weight_list[6]
        self.hidden2_w1   = weight_list[7]
        self.hidden2_w2   = weight_list[8]

    def initialize(self):
        self.set_weights([random.uniform(-0.1, +0.1) for i in range(9)])
        print("weights randomized")

    def test(self, patterns, targets):
        for pattern, target in zip(patterns, targets):
            output = self.propagate(pattern)
            if abs(target - output) <= self.tolerance:
                print(f"{pattern} --> {output}")
            else:
                print(f"{pattern} --> {output} \t(WRONG, should be {target})")
        error, correct = self.total_error(patterns, targets)
        print(f"TSS error = {error:.5f}, correct = {correct:.3f}")

    #--------------------------------------------------------------------------
    # methods to be implemented

    # returns the output produced by the network for the given pattern
    def propagate(self, pattern):
        print("not yet implemented")

    # returns a PAIR of values (e, c) where e is the total sum-squared error
    # for all patterns in the given dataset, and c is the fraction of output
    # values that are within self.tolerance of the given target values
    def total_error(self, patterns, targets):
        print("not yet implemented")

    # updates network weights and biases for the given pattern and target
    def adjust_weights(self, pattern, target):
        print("not yet implemented")

    # trains the network on all of the given patterns and targets
    # until all outputs are within tolerance of the targets
    def train(self, patterns, targets):
        print("not yet implemented")

#------------------------------------------------------------------------------

inputs = [[0,0], [0,1], [1,0], [1,1]]

ANDtargets  = [0, 0, 0, 1]
ORtargets   = [0, 1, 1, 1]
NANDtargets = [1, 1, 1, 0]
NORtargets  = [1, 0, 0, 0]
XORtargets  = [0, 1, 1, 0]  # not linearly separable
XNORtargets = [1, 0, 0, 1]  # not linearly separable

n = NeuralNetwork()

